{
    Файл является частью библиотеки RubiRobot

    © 2017-2020 Слинкин Д.А.

    Обработка двигателей контроллера Lego EV3

    Подробная информация о лицензировании находится в файле README,
    входящем в состав библиотеки.

    ДАННОЕ ПРОГРАММНОЕ ОБЕСПЕЧЕНИЕ ПРЕДОСТАВЛЯЕТСЯ «КАК ЕСТЬ», БЕЗ
    КАКИХ-ЛИБО ГАРАНТИЙ, ЯВНО ВЫРАЖЕННЫХ ИЛИ ПОДРАЗУМЕВАЕМЫХ,
    ВКЛЮЧАЯ ГАРАНТИИ ТОВАРНОЙ ПРИГОДНОСТИ, СООТВЕТСТВИЯ ПО ЕГО
    КОНКРЕТНОМУ НАЗНАЧЕНИЮ И ОТСУТСТВИЯ НАРУШЕНИЙ, НО НЕ ОГРАНИЧИВАЯСЬ 
    ИМИ. НИ В КАКОМ СЛУЧАЕ АВТОРЫ ИЛИ ПРАВООБЛАДАТЕЛИ НЕ НЕСУТ
    ОТВЕТСТВЕННОСТИ ПО КАКИМ-ЛИБО ИСКАМ, ЗА УЩЕРБ ИЛИ ПО ИНЫМ ТРЕБОВАНИЯМ,
    В ТОМ ЧИСЛЕ, ПРИ ДЕЙСТВИИ КОНТРАКТА, ДЕЛИКТЕ ИЛИ ИНОЙ СИТУАЦИИ,
    ВОЗНИКШИМ ИЗ-ЗА ИСПОЛЬЗОВАНИЯ ПРОГРАММНОГО ОБЕСПЕЧЕНИЯ ИЛИ ИНЫХ
    ДЕЙСТВИЙ С ПРОГРАММНЫМ ОБЕСПЕЧЕНИЕМ.

 **********************************************************************}


{$I uev3defines.inc}

unit rubiroMotors;


interface
uses BaseUnix,uev3Devices,uev3sysfs,uev3;
{

    run-forever: Causes the motor to run until another command is sent.
    run-to-abs-pos: Runs the motor to an absolute position specified by``position_sp`` and then stops the motor using the command specified in stop_action.
    run-to-rel-pos: Runs the motor to a position relative to the current position value. The new position will be current position + position_sp. When the new position is reached, the motor will stop using the command specified by stop_action.
    run-timed: Run the motor for the amount of time specified in time_sp and then stops the motor using the command specified by stop_action.
    run-direct: Runs the motor using the duty cycle specified by duty_cycle_sp. Unlike other run commands, changing duty_cycle_sp while running will take effect immediately.
    stop: Stop any of the run commands before they are complete using the command specified by stop_action.
    reset: Resets all of the motor parameter attributes to their default values. This will also have the effect of stopping the motor.
}

{
    running: Power is being sent to the motor.
    ramping: The motor is ramping up or down and has not yet reached a constant output level.
    holding: The motor is not turning, but rather attempting to hold a fixed position.
    overloaded: The motor is turning as fast as possible, but cannot reach its speed_sp.
    stalled: The motor is trying to run but is not turning at all.
}
const ev3motorRunning=0;
      ev3motorStoped=1;
      ev3motorHolding=2;
      ev3motorStalled=3;

type  TEv3mCommand=(cNoRun,cRunForever,cRunToAbsPos,cRunToRelPos,cRunTimed,cRunDirect,cStop,cReset);
const ev3mCommands:array[TEv3mCommand] of string[20]=('','run-forever','run-to-abs-pos','run-to-rel-pos','run-timed','run-direct','stop','reset');

type  TEv3mStopAction=(saCoast,saBrake,saHold);
const ev3mStopActions:array[TEv3mStopAction] of string[20]=('coast', 'brake', 'hold');

type  TEv3mPolarity=(pNormal,pInversed);
const ev3mPolarities:array[TEv3mPolarity] of string[20]=('normal', 'inversed');

type  TEv3mState=(sStoped,sRunning,sRamping,sHolding,sOverloaded,sStalled);
      TEv3mStates=set of TEv3mState;
{
 тип для определения и установки позиции мотора, для управления счетчиками
 cRot - вращения (обычно 360 вращений = 1 обороту, 1 вращение = 1 градусу)
 cTurn - обороты
 cDeg - градусы
}
type  TTypeCounter=(cRot,cTurn,cDeg);

const ev3mStates:array[TEv3mState]of string[20]=('','running','ramping','holding','overloaded','stalled');

// 0.2.6 Время ожидания в ms для определения факта искусственной остановки мотора.
var motorArtificalWait:dword=200;

type

  { TEv3Motor }

  TEv3Motor=class(TEv3OneDeviceHelper)
   private
    procedure resetWriteInterval;
   public
    procedure afterConstruction;override;
  end;

  { TEv3TachoMotor }
  TEv3TachoMotorClass=class of TEv3TachoMotor;
  TEv3TachoMotor=class(TEv3Motor)
    private
     frealStop:boolean;
     fcounters:array[0..9]of integer;
     fmaxSpeed:cint;
     _fspeed:double;
     _frot:double;
     fstopaction:TEv3mStopAction;
     fStopChanged:boolean;
     fpolarity:TEv3mPolarity;
     fcount_per_rot:cint;
     tmArtWait:double;
     // +++ 0.3.1
     fSuspendedCommand:TEv3mCommand;
     fSuspended:boolean;
     fGapMode:boolean;
     // --- 0.3.1
     function getcount_per_rot: integer;
     function getMaxSpeed: cint;
     procedure setFSpeed(AValue: double);
    protected
     gapCompensator:integer;
     procedure deviceRefreshed(dev:TEv3Device);override;
     property maxSpeed:cint read getMaxSpeed;
     procedure reversePolarity;
     property fspeed:double read _fspeed write setFSpeed;
     {
       Устанавливает реверс. Действует при запуске мотора.
       На текущее вращение мотора влияния не оказывает.
       После изменения реверса следует перезапустить
       все действующий на данный момент счетчики с помощью сounterStart,
       т.к. их значения приводятся в неопределенное состояние
     }
     function setReverse(yes:variant):TEv3TachoMotor;

     {
       v0.2.5 Устанавливаеет скорость мотора,
       если она раньше была не установлена, или установлена в отличное
       от fspeed значения.
       На текущее вращение мотора влияния не оказывает.
     }
     procedure setSpeed();

     function Status():TEv3mStates;

     {
       v0.3.1 Запускает или запоминает до Resume команду движения
     }
     function RunSuspend(cmd:TEv3mCommand):TEv3TachoMotor;
    public
     {
       v0.3.1 Замораживает до запуска Resume выполнение команды движения.
       Последующий вызов любого Run-метода будет не приведет в действие мотор.
       Все подготовительные действия по запуску мотора проводятся штатно,
       замораживается только сама команда.
       Между вызовами Suspend и Resume не следует вызывать более одного Run-метода.
     }
     function Suspend():TEv3TachoMotor;
     {
       v0.3.1 Выполняет команду движения, замороженную ранее с помощью Suspend
       Между вызовами Suspend и Resume не следует вызывать более одного Run-метода.
     }
     function Resume():TEv3TachoMotor;


    public
     class function varToSA(sa:variant):TEv3mStopAction;
     class function varToState(s:variant):TEv3mStates;
     // ждет события из списка событий: sStoped,sRunning,sRamping,sHolding,sOverloaded,sStalled
     function waitEvent(event:ansistring; timeout:dword=high(dword)):boolean; override;
    public
     procedure setIntervals(R,W:integer);override;
     property count_per_rot:integer read getcount_per_rot;
    public
     {
       v0.3.1 Компенсация люфта для моторов
       0,false,'no','' - компенсация отключена (значение по умолчанию)
       1,true,'yes' - компенсация включена
       При включеной компенсации, перед началом движения на заданное
       количество градусов или оборотов, если предыдущее движение
       было в обратную сторону, дополнительно сдвигает на gapCompensator
       градусов. Значение gapCompensator=3 для больших моторов
       и 0 для остальных.
       Первое движение после старта робота не может быть скомпенсировано,
       так как состояние люфта до старта двигателей предсказать невозможно.
     }

     function gapMode(yes:variant):TEv3TachoMotor;

     {
      Устанавливает режим ожидания остановки мотора.

      1. Искусственная остановка, вариант по умолчанию в RubiRobotLib.
      Если параметр realStop=false/ноль, то wait ожидает любую
      остановку мотора, в том числе и временную, когда мотор
      блокируется внешним препятствием.

      2. Естественная остановка, вариант по умолчанию в
      LEGO® MINDSTORMS® Education EV3.
      Если параметр realStop=true/не ноль, то wait ожидает
      окончательную остановку мотора.
      Таким образом, если мотор не сможет окончательно остановиться,
      например - из-за внешнего препятствия, wait зависнет.
     }
     function setWaitMode(realStop:variant):TEv3TachoMotor;

     {
      Возвращает текущий режим ожидания остановки,
      естественный (true) или искусственный (false)
     }
     function waitMode(out realStop:variant):TEv3TachoMotor;
     function waitMode:variant;

     {
      Ждет остановки мотора.
      wait без параметров использует режим ожидания остановки,
      установленный по умолчанию (искусственный)
      либо ранее примененный (wait с параметром, см. setWaitMode)
     }
     function wait(realStop:variant):TEv3TachoMotor;
     function wait():TEv3TachoMotor;


     {
      Устанавливает способ остановки мотора.
      Действует при запуске мотора.
      На текущее вращение мотора влияния не оказывает.
      Варианты:
      1) Остановка без отключения подачи электроэнергии,
         средний тормозной путь
         saBrake,'brake','тормоз','тормози','тормозить',1
      2) Остановка с удержанием позиции, минимальный тормозной путь
         saHold,'hold','стоп','стоять','стой',2
      0) Движение накатом, с отключением подачи электроэнергии,
         максимальный тормозной путь
         saCoast,0,'накат' и другие варианты, не входящие в 1) и 2)
     }
     function setHowStop(sa:variant):TEv3TachoMotor;

     {
      Возвращает способ остановки мотора
       1 - Остановка без отключения подачи электроэнергии,
           средний тормозной путь
       2 - Остановка с удержанием позиции, минимальный тормозной путь
       0 - Движение накатом, с отключением подачи электроэнергии,
           максимальный тормозной путь
     }
     function howStop(out sa:variant):TEv3TachoMotor;
     function howStop:variant;

     {
       запускает Счетчик оборотов, количество счетчиков - 10, с 0 до 9
       по умолчанию используется счетчик с номером 0
     }
     function counterStart(ID:integer=0):TEv3TachoMotor;

     {
       возвращает кол-во вращений/оборотов/градусов с момента
       запуска счетчика
       Результат будет отрицательным при обратном вращении
     }
     function counter(typeCounter:TTypeCounter=cTurn;
        ID:integer=0):variant;
     function counter(typeCounter:TTypeCounter;
        ID:integer; out count:variant):TEv3TachoMotor;

     {
       блокирует на кол-во вращений/оборотов/градусов с момента
       запуска счетчика.
       count может быть как положительным, так и отрицательным
       (позволяет контролировать реверсивное вращение мотора)
     }
     function counterWait(count:variant; typeCounter:TTypeCounter=cTurn;
        ID:integer=0):TEv3TachoMotor;

     {
       блокирует на кол-во вращений/оборотов/градусов
       с момента запуска счетчика
       count может быть только положительным
       не делает разницы между прямым и реверсивным вращениями
     }
     function counterAbsWait(count:variant; typeCounter:TTypeCounter=cTurn;
        ID:integer=0):TEv3TachoMotor;

     {
       блокирует на кол-во градусов с текущего момента
       deg (градусы) - всегда положительное значение
       не изменяет состояние мотора (не запускает и не останавливает)
       может блокировать бесконечно, если мотор не вращается.
       Для работы использует счетчик номер ID
     }
     function waitDeg(deg:integer;
        ID:integer=0):TEv3TachoMotor;

     {
       блокирует на кол-во оборотов с текущего момента
       turn (обороты) - всегда положительное значение
       не изменяет состояние мотора (не запускает и не останавливает)
       может блокировать бесконечно, если мотор не вращается.
       Для работы использует счетчик номер ID
     }
     function waitTurn(turn:double;
        ID:integer=0):TEv3TachoMotor;

     {
       пребразования вида градусы <-> вращения <-> обороты
     }
     function deg2rot(deg:variant):variant;
     function rot2deg(rot:variant):variant;
     function turn2rot(turn:variant):variant;
     function rot2turn(rot:variant):variant;

     {
       Устанавливает и возвращает текущую позицию мотора в различных
       единицах измерения
     }
     function setPosition(p:variant;
       typeCounter:TTypeCounter=cRot):TEv3TachoMotor;
     function position(out p:variant;
       typeCounter:TTypeCounter=cRot):TEv3TachoMotor;
     function position(typeCounter:TTypeCounter=cRot):variant;

     {
       Устанавливает скорость мотора в заданное значение.
       На текущее вращение мотора влияния не оказывает.
     }
     procedure setSpeed(sp:variant);
     {
       Возвращает установленную ранее скорость вращения от -100 до 100
     }
     function Speed(out sp:variant):TEv3TachoMotor;
     function Speed:variant;

     {
       Возвращает реальную скорость вращения от 0 до 100
     }
     function currentSpeed(out sp:variant):TEv3TachoMotor;
     function currentSpeed:variant;

     {
       Возвращает 1, если установлен реверс и 0 в противном случае
     }
     function Reverse(out yes:variant):TEv3TachoMotor;
     function Reverse:variant;

     {
       Останавливает мотор, способ остановки - см. setHowStop
     }
     function Stop(sa:variant):TEv3TachoMotor;
     function Stop():TEv3TachoMotor;

     { v0.2.5
       Сообщает о статусе остановки мотора
       результат:
       ev3motorRunning/0/false - мотор вращается
       ev3motorStopped/1/true - мотор остановлен
       ev3motorHolding/2/true - мотор остановлен с удержанием позиции
       ev3motorStalled/3/true - мотор временно приостановлен
                               (заблокирован внешним препятствием)
     }
     function Stoped():variant;
     function Stoped(out yes:variant):TEv3TachoMotor;

     {
       Запускает мотор (aсинхронный вызов). Возможные параметры -
       скорость и способ остановки по умолчанию при вызове stop()
     }
     function Run(sp:variant; sa:variant):TEv3TachoMotor;
     function Run(sp:variant):TEv3TachoMotor;
     function Run():TEv3TachoMotor;

     {
       Запускает мотор вплоть до достижения абсолютной позиции
       (aсинхронный вызов). Знак скорости игнорируется, движение
       направлено всегда в сторону заданной позиции.
       Позиция рассчитывается во вращениях.
       Возможные параметры - скорость и способ остановки
     }
     function RunToAbsPos(pos:variant; sp:variant; sa:variant):TEv3TachoMotor;
     function RunToAbsPos(pos:variant; sp:variant):TEv3TachoMotor;
     function RunToAbsPos(pos:variant):TEv3TachoMotor;

     {
       Запускает мотор на количество миллисекунд (aсинхронный вызов).
       Возможные параметры - скорость и способ остановки
     }
     function RunTime(ms:variant; sp:variant; sa:variant):TEv3TachoMotor;
     function RunTime(ms:variant; sp:variant):TEv3TachoMotor;
     function RunTime(ms:variant):TEv3TachoMotor;

     {
       Запускает мотор на количество градусов (aсинхронный вызов).
       Возможные параметры - скорость и способ остановки
     }
     function RunDeg(deg:variant; sp:variant; sa:variant):TEv3TachoMotor;
     function RunDeg(deg:variant; sp:variant):TEv3TachoMotor;
     function RunDeg(deg:variant):TEv3TachoMotor;

     {
       Запускает мотор на количество оборотов (aсинхронный вызов).
       Возможные параметры - скорость и способ остановки
     }
     function RunTurn(turn:variant; sp:variant; sa:variant):TEv3TachoMotor;
     function RunTurn(turn:variant; sp:variant):TEv3TachoMotor;
     function RunTurn(turn:variant):TEv3TachoMotor;

     {
       Запуск мотора с прямой подачей напряжения с блока питания
       sp - процент подаваемого напряжения, от -100 до 100
       Отрицательное значение обеспечивает инверсию вращения.
       Скорость вращения дополнительно зависит от уровня зарядки
       блока питания.
     }
     function directRun(sp:variant):TEv3TachoMotor;

     {
       Изменение напряжения, подаваемого на мотор
       (предварительно однократно вызвать directRun)
     }

     function directSpeed(sp:variant):TEv3TachoMotor;
     {
       Текущее напряжение, подаваемое на мотор
     }
     function currentDirectSpeed(out sp:variant):TEv3TachoMotor;
     function currentDirectSpeed:variant;

     {
       v0.2.5 Разгон и торможение
     }
     // устанавливает время разгона в миллисекундах
     function setRampUp(ms:variant):TEv3TachoMotor;
     // возвращает время разгона в миллисекундах
     function RampUp(out ms:variant):TEv3TachoMotor;
     function RampUp:variant;

     // устанавливает время тороможения в миллисекундах
     function setRampDown(ms:variant):TEv3TachoMotor;
     // возвращает время тороможения в миллисекундах
     function RampDown(out ms:variant):TEv3TachoMotor;
     function RampDown:variant;

     // устанавливает время разгона и торможения в миллисекундах
     function setRamp(ms:variant):TEv3TachoMotor;
     function setRamp(msUp,msDown:variant):TEv3TachoMotor;
  end;

  { TEv3MediumMotor }

  TEv3MediumMotor=class(TEv3TachoMotor)
   protected
    class function deviceType:TEv3DeviceType;override;
  end;


  { TEv3LargeMotor }

  TEv3LargeMotor=class(TEv3TachoMotor)
  protected
   class function deviceType:TEv3DeviceType;override;
  public
   constructor create(_port:variant);override;
  end;

  { TEv3Rule }

  TEv3Rule=class(TEv3)
   private
    _fspeed:double;
    faccel:double;
    fdirect:cint;
    fRampUp,fRampDown:cint;
    procedure setFSpeed(AValue: double);
   protected
    property fspeed:double read _fspeed write setFSpeed;
    // устанавливает реверс (см.TEv3TachoMotor)
    function setReverse(yes:variant):TEv3Rule;
   public
    L,R:TEv3TachoMotor;
   public
    // создание рулевого управления на моторах указанного класса
    constructor create(LPort,RPort:variant; MotorClass:TEv3TachoMotorClass);
    // создание рулевого управления на больших моторах
    constructor create(LPort,RPort:variant);
    {
     создание рулевого управления на моторах указанного класса
     порты определяются автоматически
    }
    constructor create(MotorClass:TEv3TachoMotorClass);
    {
     создание рулевого управления на больших моторах
     порты определяются автоматически
    }
    constructor create();
    destructor destroy;override;

    function Connected:boolean;
    procedure setIntervals(RR,WW:integer);

    // ждет остановки моторов
    function wait():TEv3Rule;
    function wait(realStop:variant):TEv3Rule;

    // Устанавливает способ остановки моторов (см.TEv3TachoMotor)
    function setHowStop(sa:variant):TEv3Rule;
    // Возвращает способ остановки моторов (см.TEv3TachoMotor)
    function howStop(out sa:variant):TEv3Rule;
    function howStop:variant;

    {
      Устанавливает скорость вращения от -100% до 100% (см.TEv3TachoMotor)
    }
    function setSpeed(sp:variant):TEv3Rule;
    {
      Возвращает установленную ранее скорость вращения от -100 до 100
      (см.TEv3TachoMotor)
    }
    function Speed(out sp:variant):TEv3Rule;
    function Speed:variant;

    {
      Возвращает установленную ранее скорость вращения для левого колеса
    }
    function SpeedLeft(out sp:variant):TEv3Rule;
    function SpeedLeft:variant;

    {
      Возвращает установленную ранее скорость вращения для правого колеса
    }
    function SpeedRight(out sp:variant):TEv3Rule;
    function SpeedRight:variant;

    // возвращает реверс (см.TEv3TachoMotor)
    function Reverse(out yes:variant):TEv3Rule;
    function Reverse:variant;

    // останавливает оба мотора (см.TEv3TachoMotor)
    function Stop(sa:variant):TEv3Rule;
    function Stop():TEv3Rule;

    {
      v0.3.1 Компенсация люфта для моторов
      0,false,'no','' - компенсация отключена (значение по умолчанию)
      1,true,'yes' - компенсация включена
      При включеной компенсации, перед началом движения на заданное
      количество градусов или оборотов, если предыдущее движение
      было в обратную сторону, дополнительно сдвигает на gapCompensator
      градусов. Значение gapCompensator=3 для больших моторов
      и 0 для остальных.
      Первое движение после старта робота не может быть скомпенсировано,
      так как состояние люфта до старта двигателей предсказать невозможно.
    }

    function gapMode(yes:variant):TEv3Rule;



    {
     запускает оба мотора (aсинхронный вызов) (см.TEv3TachoMotor)
     direct - направление движения
     Примеры поведения рулевого управления при положительной скорости
     и значения direct:
      100 - поворот вправо вокруг оси,
      -100 - поворот влево вокруг оси,
      50 - поворот вправо вокруг правого колеса,
      -50 - поворот влево вокруг левого колеса
    }
    function Run(direct:variant; sp:variant; sa:variant):TEv3Rule;
    function Run(direct:variant; sp:variant):TEv3Rule;
    function Run(direct:variant):TEv3Rule;

    {
      Запускает моторы на количество миллисекунд (aсинхронный вызов).
      Возможные параметры - скорость и способ остановки
    }
    function RunTime(direct:variant; ms:variant; sp:variant;
                     sa:variant):TEv3Rule;
    function RunTime(direct:variant; ms:variant; sp:variant):TEv3Rule;
    function RunTime(direct:variant; ms:variant):TEv3Rule;

    {
      Запускает моторы на количество градусов (БЛОКИРУЮЩИЙ вызов).
      Возможные параметры - скорость и способ остановки
    }
    function RunDegWait(direct:variant; deg:variant; sp:variant;
                        sa:variant):TEv3Rule;
    function RunDegWait(direct:variant; deg:variant;
                        sp:variant):TEv3Rule;
    function RunDegWait(direct:variant; deg:variant):TEv3Rule;

    {
      Запускает моторы на количество оборотов (БЛОКИРУЮЩИЙ вызов).
      Возможные параметры - скорость и способ остановки
    }
    function RunTurnWait(direct:variant; turn:variant; sp:variant;
                         sa:variant):TEv3Rule;
    function RunTurnWait(direct:variant; turn:variant;
                         sp:variant):TEv3Rule;
    function RunTurnWait(direct:variant; turn:variant):TEv3Rule;

     {
       v0.2.5 Разгон и торможение
     }
     // устанавливает на оба мотора время разгона в миллисекундах
     function setRampUp(ms:variant):TEv3Rule;
     // устанавливает на оба мотора время тороможения в миллисекундах
     function setRampDown(ms:variant):TEv3Rule;
     // устанавливает на оба мотора время разгона и торможения в миллисекундах
     function setRamp(ms:variant):TEv3Rule;
     function setRamp(msUp,msDown:variant):TEv3Rule;

     {
       v0.2.5 Тесты асинхронных
     }
     //  function RunTurn(direct:variant; turn:variant; sp:variant; sa:variant):TEv3Rule;

  end;

 {
  номерные функции возвращают двигатель по порядку подключения вне зависимости
  от номера порта. Либо nil.

  функции с параметром возвращают двигатель по номеру порта.
  идентификация портов двигателей:
  1(5)-4(8),'A'-'D', 'outA'-'outD' - вне зависимости от регистра.
 }

  function ev3Motor1:TEv3LargeMotor;
  function ev3Motor2:TEv3LargeMotor;
  function ev3Motor3:TEv3LargeMotor;
  function ev3Motor4:TEv3LargeMotor;
  function ev3Motor(index:integer=-1):TEv3LargeMotor;
  function ev3Motor(index:string):TEv3LargeMotor;
  function ev3Motor(obj:TVarRec):TEv3LargeMotor;
  function ev3Motor(obj:TObject):TEv3LargeMotor;

  function ev3MMotor1:TEv3MediumMotor;
  function ev3MMotor2:TEv3MediumMotor;
  function ev3MMotor3:TEv3MediumMotor;
  function ev3MMotor4:TEv3MediumMotor;
  function ev3MMotor(index:integer=-1):TEv3MediumMotor;
  function ev3MMotor(index:string):TEv3MediumMotor;
  function ev3MMotor(obj:TVarRec):TEv3MediumMotor;
  function ev3MMotor(obj:TObject):TEv3MediumMotor;

  function ev3Rule(index1:integer=-1; index2:integer=-1):TEv3Rule;
  function ev3Rule(index1,index2:string):TEv3Rule;
  function ev3Rule(obj:TVarRec):TEv3Rule;
  function ev3Rule(obj:TObject):TEv3Rule;

  function ev3MRule(index1:integer=-1; index2:integer=-1):TEv3Rule;
  function ev3MRule(index1,index2:string):TEv3Rule;
  function ev3MRule(obj:TVarRec):TEv3Rule;
  function ev3MRule(obj:TObject):TEv3Rule;

implementation
uses
  SysUtils,Unix,variants,dateutils;

__WARNGLOBALOFF

function ev3Motor1: TEv3LargeMotor;
begin
   result:=ev3Motor(-1);
end;

function ev3Motor2: TEv3LargeMotor;
begin
 result:=ev3Motor(-2);
end;

function ev3Motor3: TEv3LargeMotor;
begin
 result:=ev3Motor(-3);
end;

function ev3Motor4: TEv3LargeMotor;
begin
 result:=ev3Motor(-4);
end;

function ev3Motor(index: integer): TEv3LargeMotor;
begin
 if (index<0) or (index>4) then result:=ev3Helper(index,TEv3LargeMotor) as TEv3LargeMotor
 else result:=ev3Helper(index+4,TEv3LargeMotor) as TEv3LargeMotor;
end;

function ev3Motor(index: string): TEv3LargeMotor;
begin
  result:=ev3Motor(ord(TEv3DeviceHelper.varToPort(index)));
end;

function ev3Motor(obj: TVarRec): TEv3LargeMotor;
begin
  result:=ev3motor(obj.VObject);
end;

function ev3Motor(obj: TObject): TEv3LargeMotor;
begin
 result:=obj as TEv3LargeMotor;
end;

//-------
function ev3MMotor1: TEv3MediumMotor;
begin
   result:=ev3MMotor(-1);
end;

function ev3MMotor2: TEv3MediumMotor;
begin
 result:=ev3MMotor(-2);
end;

function ev3MMotor3: TEv3MediumMotor;
begin
 result:=ev3MMotor(-3);
end;

function ev3MMotor4: TEv3MediumMotor;
begin
 result:=ev3MMotor(-4);
end;

function ev3MMotor(index: integer): TEv3MediumMotor;
begin
 if (index<0) or (index>4) then result:=ev3Helper(index,TEv3MediumMotor) as TEv3MediumMotor
 else result:=ev3Helper(index+4,TEv3MediumMotor) as TEv3MediumMotor;
end;

function ev3MMotor(index: string): TEv3MediumMotor;
begin
 result:=ev3MMotor(ord(TEv3DeviceHelper.varToPort(index)));
end;

function ev3MMotor(obj: TVarRec): TEv3MediumMotor;
begin
 result:=ev3MMotor(obj.VObject);
end;

function ev3MMotor(obj: TObject): TEv3MediumMotor;
begin
 result:=obj as TEv3MediumMotor;
end;


// -----------------

var
ev3FirstRule:TEv3Rule=nil;
ev3Rules:array[outA..outD,outA..outD]of Tev3Rule=(
 (nil,nil,nil,nil),
 (nil,nil,nil,nil),
 (nil,nil,nil,nil),
 (nil,nil,nil,nil)
);

function ev3Rule(index1: integer; index2: integer): TEv3Rule;
 var lp,rp:TEv3Port;
begin
  if index1*index2<0 then exit(nil); // неодинаковый способ инициализации моторов запрещен
  if index1<0 then begin
   if assigned(ev3FirstRule) then exit(ev3FirstRule);
   lp:=ev3Motor1.port;
   if not (lp in [outA..outD]) then exit(nil);
   rp:=ev3Motor2.port;
   if not (rp in [outA..outD]) then exit(nil);
   ev3FirstRule:=ev3rule(ord(lp),ord(rp));
   exit(ev3FirstRule);
  end;
  // теперь - по переданным портам
  if index1<5 then inc(index1,4);
  if index2<5 then inc(index2,4);
  lp:=TEv3Port(index1);
  rp:=TEv3Port(index2);
  if assigned(ev3Rules[lp,rp]) then exit(ev3Rules[lp,rp]);
  ev3Rules[lp,rp]:=TEv3Rule.Create(lp,rp);
  result:=ev3Rules[lp,rp];
end;

function ev3Rule(index1, index2: string): TEv3Rule;
begin
 result:=ev3Rule(ord(TEv3DeviceHelper.varToPort(index1)),ord(TEv3DeviceHelper.varToPort(index2)));
end;

function ev3Rule(obj: TVarRec): TEv3Rule;
begin
 result:=ev3Rule(obj.VObject);
end;

function ev3Rule(obj: TObject): TEv3Rule;
begin
 result:=obj as TEv3Rule;
end;

//-----------------
var
ev3MFirstRule:TEv3Rule=nil;
ev3MRules:array[outA..outD,outA..outD]of Tev3Rule=(
 (nil,nil,nil,nil),
 (nil,nil,nil,nil),
 (nil,nil,nil,nil),
 (nil,nil,nil,nil)
);

function ev3MRule(index1: integer; index2: integer): TEv3Rule;
var lp,rp:TEv3Port;
begin
 if index1*index2<0 then exit(nil); // неодинаковый способ инициализации моторов запрещен
 if index1<0 then begin
  if assigned(ev3FirstRule) then exit(ev3FirstRule);
  lp:=ev3MMotor1.port;
  if not (lp in [outA..outD]) then exit(nil);
  rp:=ev3MMotor2.port;
  if not (rp in [outA..outD]) then exit(nil);
  ev3MFirstRule:=ev3Mrule(ord(lp),ord(rp));
  exit(ev3MFirstRule);
 end;
 // теперь - по переданным портам
 if index1<5 then inc(index1,4);
 if index2<5 then inc(index2,4);
 lp:=TEv3Port(index1);
 rp:=TEv3Port(index2);
 if assigned(ev3MRules[lp,rp]) then exit(ev3MRules[lp,rp]);
 ev3MRules[lp,rp]:=TEv3Rule.Create(lp,rp,TEv3MediumMotor);
 result:=ev3MRules[lp,rp];
end;

function ev3MRule(index1, index2: string): TEv3Rule;
begin
 result:=ev3MRule(ord(TEv3DeviceHelper.varToPort(index1)),ord(TEv3DeviceHelper.varToPort(index2)));
end;

function ev3MRule(obj: TVarRec): TEv3Rule;
begin
 result:=ev3MRule(obj.VObject);
end;

function ev3MRule(obj: TObject): TEv3Rule;
begin
 result:=obj as TEv3Rule;
end;

{ TEv3LargeMotor }

class function TEv3LargeMotor.deviceType: TEv3DeviceType;
begin
 Result:=ev3tLTachoMotor;
end;

constructor TEv3LargeMotor.create(_port: variant);
begin
  inherited create(_port);
  gapCompensator:=3;
end;


{ TEv3Motor }

procedure TEv3Motor.resetWriteInterval;
 var i:TEv3DeviceAttribute;
begin
  for i in TEv3DeviceAttribute do
   if assigned(ev.Attrs[i]) then ev.Attrs[i].skipInterval:=true;
end;

procedure TEv3Motor.afterConstruction;
begin
  resetWriteInterval;
end;

//procedure TEv3Motor.setIntervals(R, W: integer);
//begin
  // -----
//end;


{ TEv3Rule }

constructor TEv3Rule.create(LPort, RPort: variant;
  MotorClass: TEv3TachoMotorClass);
begin
  inherited create;
  L:=ev3Helper(ord(TEv3DeviceHelper.varToPort(lport)),MotorClass) as TEv3TachoMotor;
  if not assigned(L) then fail;
  R:=ev3Helper(ord(TEv3DeviceHelper.varToPort(rport)),MotorClass) as TEv3TachoMotor;
  if not assigned(R) then fail;
end;

constructor TEv3Rule.create(LPort, RPort: variant);
begin
 create(LPort, RPort, TEv3LargeMotor);
end;

constructor TEv3Rule.create(MotorClass: TEv3TachoMotorClass);
begin
  inherited create;
  L:=ev3Helper(-1,MotorClass) as TEv3TachoMotor;
  if not assigned(L) then fail;
  R:=ev3Helper(-2,MotorClass) as TEv3TachoMotor;
  if not assigned(R) then begin L.free; fail; end;
end;

constructor TEv3Rule.create();
begin
 create(TEv3LargeMotor);
end;


destructor TEv3Rule.destroy;
begin
  inherited destroy;
end;

function TEv3Rule.setHowStop(sa: variant): TEv3Rule;
begin
  result:=self;
  L.setHowStop(sa); R.setHowStop(sa);
end;

function TEv3Rule.howStop(out sa: variant): TEv3Rule;
begin
 result:=self;
 L.HowStop(sa);
end;

function TEv3Rule.howStop: variant;
begin
 result:=L.howStop;
end;

function TEv3Rule.setSpeed(sp: variant): TEv3Rule;
 var dspeed:double;
begin
 result:=self;
 fspeed:=sp;
 faccel:=1;
 if fdirect<-100 then fdirect:=-100;
 if fdirect>100 then fdirect:=100;
 if fdirect=0 then begin
   L.fspeed:=_fspeed; R.fspeed:=_fspeed;
//   L.setSpeed(_fspeed); R.setSpeed(_fspeed);

   l.setRampUp(fRampUp); L.setRampDown(fRampDown);
   R.setRampUp(fRampUp); R.setRampDown(fRampDown);
   exit;
 end;

 dspeed:=fdirect*_fspeed/50;
 if fdirect<0 then begin
   L.fspeed:=_fspeed+dspeed; R.fspeed:=_fspeed;
//   L.setSpeed(_fspeed+dspeed); R.setSpeed(_fspeed);
   if _fspeed=0 then exit;
   faccel:=abs((_fspeed+dspeed)/_fspeed);
   r.setRampUp(fRampUp*faccel);r.setRampDown(fRampDown*faccel);
   l.setRampUp(fRampUp);l.setRampDown(fRampDown);
 end
 else begin
   L.fspeed:=_fspeed; R.fspeed:=_fspeed-dspeed;
 //  WD(l.fspeed,' ',R.fspeed);
   //   L.setSpeed(_fspeed); R.setSpeed(_fspeed-dspeed);
   if _fspeed=0 then exit;
   faccel:=abs((_fspeed-dspeed)/_fspeed);
   r.setRampUp(fRampUp);r.setRampDown(fRampDown);
   l.setRampUp(fRampUp*faccel);l.setRampDown(fRampDown*faccel);
 end;
end;

function TEv3Rule.Speed(out sp: variant): TEv3Rule;
begin
 result:=self;
 sp:=self.Speed;
end;

function TEv3Rule.Speed: variant;
begin
  result:=fspeed;
end;

function TEv3Rule.SpeedLeft(out sp: variant): TEv3Rule;
begin
 result:=self;
 L.Speed(sp);
end;

function TEv3Rule.SpeedLeft: variant;
begin
 result:=L.speed;
end;

function TEv3Rule.SpeedRight(out sp: variant): TEv3Rule;
begin
 result:=self;
 R.Speed(sp);
end;

function TEv3Rule.SpeedRight: variant;
begin
 result:=R.speed;
end;


function TEv3Rule.setReverse(yes: variant): TEv3Rule;
begin
 result:=self;
 L.setReverse(yes);
 R.setReverse(yes);
end;

function TEv3Rule.Reverse(out yes: variant): TEv3Rule;
begin
 result:=self;
 L.Reverse(yes);
end;

function TEv3Rule.Reverse: variant;
begin
 result:=L.Reverse;
end;

procedure TEv3Rule.setFSpeed(AValue: double);
begin
  if avalue>100 then avalue:=100
  else
  if avalue<-100 then avalue:=-100;
  if _fspeed=AValue then Exit;
  _fspeed:=AValue;
end;

function TEv3Rule.Connected: boolean;
begin
  if not assigned(self) then exit(false);
  result:=l.Connected and r.connected;
end;

procedure TEv3Rule.setIntervals(RR, WW: integer);
begin
  L.setIntervals(RR,WW);
  R.setIntervals(RR,WW);
end;

function TEv3Rule.wait(): TEv3Rule;
begin
  result:=self;
  if not L.fRealStop then begin
   L._timerStart(L.tmArtWait);
   R._timerStart(R.tmArtWait);
   waitAnyEvent(
     [L,sStoped,L,sHolding,L,sStalled,R,sStoped,R,sHolding,R,sStalled]
     );
  end
  else begin
   waitAnyEvent(
   [L,sStoped,L,sHolding,R,sStoped,R,sHolding]
   );
  end;
  L.stop();
  R.Stop();
end;

function TEv3Rule.wait(realStop: variant): TEv3Rule;
begin
  result:=self;
  R.frealStop:=realStop;
  L.frealStop:=realStop;
  self.wait;
end;


function TEv3Rule.Stop(sa: variant): TEv3Rule;
begin
 result:=self;
 L.Stop(sa);
 R.Stop(sa);
end;

function TEv3Rule.Stop(): TEv3Rule;
begin
 result:=self;
 L.Stop();
 R.Stop();
end;

function TEv3Rule.gapMode(yes: variant): TEv3Rule;
begin
  result:=self;
  l.gapMode(yes);
  l.gapMode(yes);
end;

function TEv3Rule.Run(direct: variant; sp: variant; sa: variant): TEv3Rule;
begin
  setHowStop(sa);
  result:=self.run(direct,sp);
end;

function TEv3Rule.Run(direct: variant; sp: variant): TEv3Rule;
begin
 result:=self;
 fdirect:=direct;
 setSpeed(sp);
 L.Run;
 R.Run;
end;

function TEv3Rule.Run(direct: variant): TEv3Rule;
begin
 result:=self.Run(direct,speed);
end;

function TEv3Rule.RunTime(direct: variant; ms: variant; sp: variant; sa: variant
  ): TEv3Rule;
begin
 setHowStop(sa);
 result:=self.RunTime(direct,ms,sp);
end;

function TEv3Rule.RunTime(direct: variant; ms: variant; sp: variant): TEv3Rule;
begin
 result:=self;
 fdirect:=direct;
 setSpeed(sp);
 L.RunTime(ms);
 R.RunTime(ms);
end;

function TEv3Rule.RunTime(direct: variant; ms: variant): TEv3Rule;
begin
 result:=self.RunTime(direct,ms,speed);
end;

function TEv3Rule.RunDegWait(direct: variant; deg: variant; sp: variant; sa: variant
  ): TEv3Rule;
begin
 setHowStop(sa);
 result:=self.RunDegWait(direct,deg,sp);
end;

function TEv3Rule.RunDegWait(direct: variant; deg: variant; sp: variant): TEv3Rule;
begin
 result:=RunTurnWait(direct,double(deg)/360.0,sp);
end;

function TEv3Rule.RunDegWait(direct: variant; deg: variant): TEv3Rule;
begin
 result:=self.RunDegWait(direct,deg,speed);
end;

function TEv3Rule.RunTurnWait(direct: variant; turn: variant; sp: variant;
  sa: variant): TEv3Rule;
begin
 setHowStop(sa);
 result:=self.RunTurnWait(direct,turn,sp);
end;

function TEv3Rule.RunTurnWait(direct: variant; turn: variant; sp: variant
  ): TEv3Rule;
begin
 result:=self;
 fdirect:=direct;
 setSpeed(sp);
 if fdirect=0 then begin
// +++ 0.3.1
// остановка любого мотора должна служить причиной остановки всей конструкции
//  R.RunTurn(turn);
//  L.RunTurn(turn).wait();
//  R.wait();
  R.RunTurn(turn);
  L.RunTurn(turn);
  while ((r.Stoped()=0) and (l.stoped()=0)) do;
  l.stop; r.stop;
// --- 0.3.1
 end else
 if fdirect>0 then begin
  if turn<0 then R.setFSpeed(-R.fspeed);
  R.Run;
  L.RunTurn(turn).wait();
  R.Stop;
 end else begin
  if turn<0 then L.setFSpeed(-L.fspeed);
  L.Run;
  R.RunTurn(turn).wait();
  L.Stop;
 end;
end;

function TEv3Rule.RunTurnWait(direct: variant; turn: variant): TEv3Rule;
begin
 result:=self;
 self.RunTurnWait(direct,Turn,speed);
end;

{
begin
 result:=self;
 setHowStop(sa);
 fdirect:=direct;
 setSpeed(sp);
 if direct>0 then begin
  if turn<0 then R.setFSpeed(-R.fspeed);
  L.RunTurn(turn);
  R.RunTurn(double(turn)*faccel);
 end else begin
  if turn<0 then L.setFSpeed(-L.fspeed);
  L.RunTurn(double(turn)*faccel);
  R.RunTurn(turn);
 end;
end;
}

function TEv3Rule.setRampUp(ms: variant): TEv3Rule;
begin
 result:=self;
 fRampUp:=ms;
end;

function TEv3Rule.setRampDown(ms: variant): TEv3Rule;
begin
 result:=self;
 fRampUp:=ms;
end;

function TEv3Rule.setRamp(ms: variant): TEv3Rule;
begin
 result:=self.setRamp(ms,ms);
end;

function TEv3Rule.setRamp(msUp, msDown: variant): TEv3Rule;
begin
 result:=self;
 setRampUp(msUp);
 setRampDown(msDown);
end;

{ TEv3MediumMotor }

class function TEv3MediumMotor.deviceType: TEv3DeviceType;
begin
  Result:=ev3tMTachoMotor;
end;


{ TEv3TachoMotor }

function TEv3TachoMotor.getMaxSpeed: cint;
begin
  if fmaxspeed=0 then fmaxspeed:=ev.Attrs[ev3amax_speed].iData;
  result:=fmaxspeed;
end;

function TEv3TachoMotor.getcount_per_rot: integer;
begin
  if fcount_per_rot=0 then fcount_per_rot:=ev.Attrs[ev3acount_per_rot].iData;
  exit(fcount_per_rot);
end;


procedure TEv3TachoMotor.setFSpeed(AValue: double);
begin
  if avalue>100 then avalue:=100
  else
  if avalue<-100 then avalue:=-100;
  if _fspeed=AValue then Exit;
  _fspeed:=AValue;
end;

procedure TEv3TachoMotor.deviceRefreshed(dev: TEv3Device);
begin
  inherited deviceRefreshed(dev);
  dev.Attrs[ev3acommand].sData:=ev3mCommands[cReset];
  dev.Attrs[ev3astop_action].sData:=ev3mStopActions[fstopaction];
  dev.Attrs[ev3apolarity].sData:=ev3mPolarities[fpolarity];
end;

procedure TEv3TachoMotor.reversePolarity;
begin
  if reverse then setReverse(0) else setReverse(1);
end;


class function TEv3TachoMotor.varToSA(sa: variant): TEv3mStopAction;
begin
 case utf8lowercase(trim(varToStr(sa))) of
  'brake','тормоз','тормози','тормозить','1':result:=saBrake;
  'hold','стоп','стоять','стой','2':result:=saHold
  else result:=saCoast;
 end;
end;


procedure outSet(s:TEv3mStates);
 var i:TEv3mState;
begin
 for i in s do write(i,' ');
 writeln;
end;

class function TEv3TachoMotor.varToState(s: variant): TEv3mStates;
 var us:string;
     st:TEv3mState;
begin
 us:=utf8lowercase(trim(varToStr(s)));
 result:=[];
 // sStoped,sRunning,sRamping,sHolding,sOverloaded,sStalled
 for st in TEv3mState do begin
  // writeln(us,' == ',ev3mStates[st],' +- ', pos(ev3mStates[st],us)<>0 );
   if (pos(ev3mStates[st],us)<>0)
   //(ev3mStates[st]=us)
   or(us=inttostr(ord(st))) then include(result,st);
 end;
 case us of
  'работа':include(result,sRunning);
  'разгон':include(result,sRamping);
  'стоп':include(result,sHolding);
  'перегрузка':include(result,sOverloaded);
  'блок':include(result,sStalled);
  //else include(result,sStoped);
 end;
 if not (sRunning in result) then include(result,sStoped);
end;



function TEv3TachoMotor.waitEvent(event: ansistring; timeout: dword): boolean;
var st:TEv3mStates;
    s:string; tm:double;
    //cntArtWait:integer;
begin
 st:=varToState(event);
 _timerStart(tm);
 while true do begin
  s:=ev.Attrs[ev3astate].sData;
  if varToState(s)*st<>[] then begin
    if not (sStalled in st) then exit(true);
    // ожидали и получили блокировку,
    // проверяем ее на непрерывность за период motorArtificalWait
    if _timer(tmArtWait)>motorArtificalWait then exit(true);
  end
  else begin
   if (sStalled in st) then begin
    // ожидали, но не получили блокировку, обнуляем таймер
     _timerStart(tmArtWait);
   end;
  end;
  if _timer(tm)>timeout then exit(false);
  sleep(waitInterval);
 end;
end;

function TEv3TachoMotor.counterStart(ID: integer): TEv3TachoMotor;
begin
  result:=self;
  fcounters[id]:=position;
end;

function TEv3TachoMotor.counter(typeCounter: TTypeCounter; ID: integer
  ): variant;
begin
   case typeCounter of
    cTurn: exit(rot2turn(counter(cRot,id)));
    cDeg: exit(rot2deg(counter(cRot,id)));
   end;
   result:=position-fcounters[id];
end;

function TEv3TachoMotor.counter(typeCounter: TTypeCounter; ID: integer; out
  count: variant): TEv3TachoMotor;
begin
   result:=self;
   count:=counter(typeCounter,id)
end;


function TEv3TachoMotor.counterWait(count: variant; typeCounter: TTypeCounter;
  ID: integer): TEv3TachoMotor;
begin
  result:=self;
  while double(counter(typeCounter,id))<count do;
end;

function TEv3TachoMotor.counterAbsWait(count: variant;
  typeCounter: TTypeCounter; ID: integer): TEv3TachoMotor;
begin
 result:=self;
 while abs(double(counter(typeCounter,id)))<count do;
end;

function TEv3TachoMotor.waitDeg(deg: integer; ID: integer): TEv3TachoMotor;
begin
 result:=counterStart(ID).counterAbsWait(deg,cdeg,ID);
end;

function TEv3TachoMotor.waitTurn(turn: double; ID: integer): TEv3TachoMotor;
begin
 result:=counterStart(ID).counterAbsWait(turn,cTurn,ID);
end;

function TEv3TachoMotor.deg2rot(deg: variant): variant;
begin
 result:=round(double(deg)/360*count_per_rot);
end;

function TEv3TachoMotor.rot2deg(rot: variant): variant;
begin
 result:=round(double(rot)/count_per_rot*360);
end;

function TEv3TachoMotor.turn2rot(turn: variant): variant;
begin
 result:=round(double(turn)*count_per_rot);
end;

function TEv3TachoMotor.rot2turn(rot: variant): variant;
begin
 result:=double(rot)/count_per_rot;
end;

procedure TEv3TachoMotor.setIntervals(R, W: integer);
begin
 ev.Attrs[ev3astop_action].setIntervals(R,W);
 ev.Attrs[ev3aposition].setIntervals(R,W);
 ev.Attrs[ev3aspeed_sp].setIntervals(R,W);
 ev.Attrs[ev3aspeed].setIntervals(R,W);
 ev.Attrs[ev3apolarity].setIntervals(R,W);
 ev.Attrs[ev3acommand].setIntervals(R,W);
 ev.Attrs[ev3atime_sp].setIntervals(R,W);
 ev.Attrs[ev3aposition_sp].setIntervals(R,W);
end;

function TEv3TachoMotor.gapMode(yes: variant): TEv3TachoMotor;
begin
  result:=self;
  fGapMode:=varToYes(yes);
end;

function TEv3TachoMotor.wait(realStop: variant): TEv3TachoMotor;
begin
 result:=self;
 frealStop:=realstop;
 self.wait;
end;

function TEv3TachoMotor.wait(): TEv3TachoMotor;
begin
   result:=self;
   if not fRealStop then begin
    _timerStart(tmArtWait);
    waitAnyEvent([self,sStoped,self,sHolding,self,sStalled]);
    stop();
   end
   else
    waitAnyEvent([self,sStoped,self,sHolding]);
end;

function TEv3TachoMotor.setHowStop(sa: variant): TEv3TachoMotor;
begin
  result:=self;
  fstopaction:=varToSA(sa);
  if ev.Attrs[ev3astop_action].sStored=ev3mStopActions[fStopAction] then exit;
  ev.Attrs[ev3astop_action].sData:=ev3mStopActions[fStopAction];
end;

function TEv3TachoMotor.howStop(out sa: variant): TEv3TachoMotor;
begin
 result:=self;
 sa:=self.howStop;
end;

function TEv3TachoMotor.howStop: variant;
begin
 with ev.Attrs[ev3astop_action] do begin
  if isStored then result:=varToSA(sStored)
  else result:=varToSA(sData);
 end;
end;

function TEv3TachoMotor.setPosition(p: variant; typeCounter: TTypeCounter
  ): TEv3TachoMotor;
begin
 result:=self;
 case typeCounter of
    cTurn: ev.Attrs[ev3aposition].iData:=turn2rot(p);
    cDeg: ev.Attrs[ev3aposition].iData:=deg2rot(p);
    else ev.Attrs[ev3aposition].iData:=Round(double(p));
 end;
end;

function TEv3TachoMotor.position(out p: variant; typeCounter: TTypeCounter
  ): TEv3TachoMotor;
begin
 result:=self;
 p:=position(typeCounter);
end;

function TEv3TachoMotor.position(typeCounter: TTypeCounter): variant;
 var mul:integer;
begin
 if reverse then mul:=-1 else mul:=1;
 case typeCounter of
    cTurn: result:=rot2turn(ev.Attrs[ev3aposition].iData)*mul;
    cDeg: result:=rot2deg(ev.Attrs[ev3aposition].iData)*mul;
    else result:=ev.Attrs[ev3aposition].iData*mul;
 end;
end;

{
function TEv3TachoMotor.setSpeed(sp: variant): TEv3TachoMotor;
begin
 result:=self;
 fspeed:=sp;
 if integer(sp)<0 then setReverse(1) else setReverse(0);
 setSpeed();
end;
}

function TEv3TachoMotor.Speed(out sp: variant): TEv3TachoMotor;
begin
  result:=self;
  sp:=self.speed;
end;

function TEv3TachoMotor.Speed: variant;
begin
 result:=fspeed;
end;

function TEv3TachoMotor.currentSpeed(out sp: variant): TEv3TachoMotor;
begin
 result:=self;
 sp:=self.currentSpeed;
end;

function TEv3TachoMotor.currentSpeed: variant;
begin
 result:=(ev.Attrs[ev3aspeed].iData*100)/double(maxSpeed);
end;

function TEv3TachoMotor.setReverse(yes: variant): TEv3TachoMotor;
 var s:string;
begin
 result:=self;
// writeln('Reverse = ',TEv3mPolarity(varToYes(yes)));
 s:=ev3mPolarities[TEv3mPolarity(varToYes(yes))];
 //writeln(s);
 with ev.Attrs[ev3apolarity] do begin
  if isStored and (sStored=s) then exit;
  sData:=s;
 end;
end;

procedure TEv3TachoMotor.setSpeed();
 var i:longint;
begin
 i:=round(fspeed*maxSpeed/100);
 with ev.Attrs[ev3aspeed_sp] do begin
  if (not isStored) or (iStored<>i) then idata:=i;
 end;
end;

procedure TEv3TachoMotor.setSpeed(sp: variant);
begin
  fspeed:=sp;
  setSpeed();
end;

function TEv3TachoMotor.Reverse(out yes: variant): TEv3TachoMotor;
begin
 result:=self;
 yes:=self.reverse;
end;

function TEv3TachoMotor.Reverse: variant;
 var s:string;
begin
 with ev.Attrs[ev3apolarity] do begin
  if isStored then s:=sStored
  else s:=sData;
 end;
 result:=ev3mPolarities[pInversed]=s;
end;

function TEv3TachoMotor.Stop(sa: variant): TEv3TachoMotor;
 var tmp:TEv3mStopAction;
begin
 result:=self;
 tmp:=varToSA(sa);
 if tmp<>fstopaction then fStopChanged:=true;
 setHowStop(sa);
 self.stop;
end;

function TEv3TachoMotor.Stop(): TEv3TachoMotor;
 var st:TEv3mStates; vsa:TEv3mState;
begin
  result:=self;
  //writeln(fstopaction);

  st:=varToState(ev.Attrs[ev3astate].sData);

  // повторная остановка с удержанием, когда мотор уже стоит с удержанием,
  // даст небольшой сдвиг вперед,
  // поэтому блокируем сие безосмысленно-вредоносное действо
  if (fstopaction=saHold) and (sHolding in st) then exit;

  // используется команда перемещения отн. позицию вместо команды остановки
  // это нужно для возможности сменить способ остановки, т.к. он игнорируется при команде остановки.
  // при этом перемещение в относит. позицию 0 косвенно дает остановку
  if fStopChanged then begin
    ev.Attrs[ev3astop_action].sData:=ev3mStopActions[fstopaction];
    ev.Attrs[ev3aposition_sp].iData:=0;
    ev.Attrs[ev3acommand].sData:=ev3mCommands[cRunToRelPos];
    fStopChanged:=false;
    // ожидаем реальной остановки/блокировки
    if fstopaction=saHold then vsa:=sHolding else vsa:=sStoped;
    while true do begin
      st:=varToState(ev.Attrs[ev3astate].sData);
      if [vsa,sStalled]*st<>[] then break;
    end;
  end
  else
   // если уже стоит
   if sStoped in st then exit
   // однако если последний Stop не менял способ остановки - используем именно команду остановки
   else ev.Attrs[ev3acommand].sData:=ev3mCommands[cStop];
end;

function TEv3TachoMotor.Stoped(): variant;
 var st:TEv3mStates;
begin
  st:=status();
  //outSet(st);
  if sStoped in st then exit(ev3MotorStoped);
  if sStalled in st then exit(ev3MotorStalled);
  if sHolding in st then exit(ev3MotorHolding);
  result:=ev3MotorRunning;
end;

function TEv3TachoMotor.Stoped(out yes: variant): TEv3TachoMotor;
begin
  result:=self;
  yes:=self.Stoped;
end;

function TEv3TachoMotor.waitMode(out realStop: variant): TEv3TachoMotor;
begin
 result:=self;
 realStop:=self.waitMode;
end;

function TEv3TachoMotor.waitMode: variant;
begin
 result:=fRealStop;
end;

function TEv3TachoMotor.setWaitMode(realStop: variant): TEv3TachoMotor;
begin
 result:=self;
 fRealStop:=realStop;
end;

function TEv3TachoMotor.Status(): TEv3mStates;
begin
 result:=varToState(ev.Attrs[ev3astate].sData);
end;

function TEv3TachoMotor.Suspend(): TEv3TachoMotor;
begin
  result:=self;
  fSuspended:=true;
end;

function TEv3TachoMotor.Resume(): TEv3TachoMotor;
begin
 result:=self;
 if not fSuspended then exit;
 if fSuspendedCommand=cNoRun then exit;
 fSuspended:=false;
 RunSuspend(fSuspendedCommand);
 fSuspendedCommand:=cNoRun;
end;

function TEv3TachoMotor.RunSuspend(cmd: TEv3mCommand): TEv3TachoMotor;
begin
 result:=self;
 if fSuspended then fSuspendedCommand:=cmd
 else ev.Attrs[ev3acommand].sData:=ev3mCommands[cmd];
end;

function TEv3TachoMotor.Run(sp: variant; sa: variant): TEv3TachoMotor;
begin
  setHowStop(sa);
  result:=self.Run(sp);
end;

function TEv3TachoMotor.Run(sp: variant): TEv3TachoMotor;
begin
 //+++0.2.5.1
 if stoped=3 then stop(); // если заблокирован, остановить
 //---0.2.5.1
 result:=self;
 fspeed:=sp;
 setReverse(0);
 setSpeed();
 runSuspend(cRunForever);
end;

function TEv3TachoMotor.Run(): TEv3TachoMotor;
begin
 result:=self;
 self.Run(fspeed);
end;

function TEv3TachoMotor.RunToAbsPos(pos: variant; sp: variant; sa: variant
  ): TEv3TachoMotor;
begin
 setHowStop(sa);
 result:=self.runToAbsPos(pos,sp);
end;

function TEv3TachoMotor.RunToAbsPos(pos: variant; sp: variant): TEv3TachoMotor;
begin
 //+++0.2.5.1
 if stoped=3 then stop(); // если заблокирован, остановить
 //---0.2.5.1
 result:=self;
 fspeed:=sp;
 setReverse(0);
 setSpeed();
 ev.Attrs[ev3aposition_sp].sData:=pos;
 runSuspend(cRunToAbsPos);
end;

function TEv3TachoMotor.RunToAbsPos(pos: variant): TEv3TachoMotor;
begin
 result:=self.RunToAbsPos(pos,fspeed);
end;

function TEv3TachoMotor.RunTime(ms: variant; sp: variant; sa: variant
  ): TEv3TachoMotor;
begin
  setHowStop(sa);
  result:=self.runTime(ms,sp);
end;

function TEv3TachoMotor.RunTime(ms: variant; sp: variant): TEv3TachoMotor;
begin
 //+++0.2.5.1
 if stoped=3 then stop(); // если заблокирован, остановить
 //---0.2.5.1
 //+++0.2.6
 //while stoped=3 do sleep(10);
 //---0.2.6

 result:=self;
 fspeed:=sp;
 setReverse(0);
 setSpeed();
 ev.Attrs[ev3atime_sp].sData:=ms;
 runSuspend(cRunTimed);
end;

function TEv3TachoMotor.RunTime(ms: variant): TEv3TachoMotor;
begin
 result:=self.runtime(ms,fspeed);
end;

function TEv3TachoMotor.RunDeg(deg: variant; sp: variant; sa: variant
  ): TEv3TachoMotor;
begin
 setHowStop(sa);
 result:=self.rundeg(deg,sp);
end;

function TEv3TachoMotor.RunDeg(deg: variant; sp: variant): TEv3TachoMotor;
begin
  result:=RunTurn(deg/360.0,sp);
end;

function TEv3TachoMotor.RunDeg(deg: variant): TEv3TachoMotor;
begin
  result:=RunTurn(double(deg)/360.0);
end;

function TEv3TachoMotor.RunTurn(turn: variant; sp: variant; sa: variant
  ): TEv3TachoMotor;
begin
 setHowStop(sa);
 result:=RunTurn(turn,sp);
end;

{
function TEv3TachoMotor.RunTurn(turn: variant; sp: variant): TEv3TachoMotor;
begin
 result:=self;
 if fcount_per_rot=0 then fcount_per_rot:=ev.Attrs[ev3acount_per_rot].iData;
 setSpeed(sp); // обязательно так, ибо возможно нужен будет реверс
// fspeed:=sp;
// setSpeed();
// writeln('turn - ',Round(double(turn)*fcount_per_rot));
 ev.Attrs[ev3aposition_sp].iData:=turn2rot(turn);
 ev.Attrs[ev3acommand].sData:=ev3mCommands[cRunToRelPos];
end;
}

function TEv3TachoMotor.RunTurn(turn: variant; sp: variant): TEv3TachoMotor;
 var _sp,_prevsp,_rot:double;
     rot:variant;
begin
 _sp:=toReal(sp);
 _prevsp:=fspeed;
 //+++0.2.5.1
 if stoped=3 then stop(); // если заблокирован, остановить
 //---0.2.5.1
 result:=self;
 if fcount_per_rot=0 then fcount_per_rot:=ev.Attrs[ev3acount_per_rot].iData;
 fspeed:=_sp;
 setReverse(0);
 setSpeed();
 _rot:=turn2rot(turn);
 if fspeed<0 then turn:=-turn; // т.к. при cRunToRelPos знак скорости игнорируется
 rot:=turn2rot(turn);

 if fGapMode and (_sp*_prevsp*_frot*_rot<0) then begin
   if turn<0 then rot-=deg2rot(gapCompensator)
   else rot+=deg2rot(gapCompensator);
   writeln(rot);
 end;
 ev.Attrs[ev3aposition_sp].iData:=rot;
 _frot:=_rot;
 runSuspend(cRunToRelPos);
end;



function TEv3TachoMotor.RunTurn(turn: variant): TEv3TachoMotor;
begin
  result:=self;
  RunTurn(turn,fspeed);
end;


function TEv3TachoMotor.directRun(sp: variant): TEv3TachoMotor;
begin
 result:=directSpeed(sp);
 runSuspend(cRunDirect);
end;

function TEv3TachoMotor.directSpeed(sp: variant): TEv3TachoMotor;
begin
 result:=self;
 if sp>100 then sp:=100;
 if sp<-100 then sp:=-100;
 ev.Attrs[ev3aduty_cycle_sp].iData:=integer(sp);
end;

function TEv3TachoMotor.currentDirectSpeed(out sp: variant): TEv3TachoMotor;
begin
 result:=self;
 sp:=currentDirectSpeed();
end;

function TEv3TachoMotor.currentDirectSpeed: variant;
begin
 result:=ev.Attrs[ev3aduty_cycle_sp].iData;
end;

function TEv3TachoMotor.setRampUp(ms: variant): TEv3TachoMotor;
begin
 result:=self;
 with ev.Attrs[ev3aramp_up_sp] do begin
   if isStored and (iStored=ms) then exit;
   iData:=ms;
 end;
end;

function TEv3TachoMotor.RampUp(out ms: variant): TEv3TachoMotor;
begin
 result:=self;
 ms:=self.RampUp;
end;

function TEv3TachoMotor.RampUp: variant;
begin
 with ev.Attrs[ev3aramp_up_sp] do begin
   if isStored then result:=iStored
   else result:=iData;
 end;
end;

function TEv3TachoMotor.setRampDown(ms: variant): TEv3TachoMotor;
begin
 result:=self;
 with ev.Attrs[ev3aramp_down_sp] do begin
   if isStored and (iStored=ms) then exit;
   iData:=ms;
 end;
end;

function TEv3TachoMotor.RampDown(out ms: variant): TEv3TachoMotor;
begin
 result:=self;
 ms:=self.RampDown;
end;

function TEv3TachoMotor.RampDown: variant;
begin
 with ev.Attrs[ev3aramp_down_sp] do begin
   if isStored then result:=iStored
   else result:=iData;
 end;
end;

function TEv3TachoMotor.setRamp(ms: variant): TEv3TachoMotor;
begin
  result:=self.setRamp(ms,ms);
end;

function TEv3TachoMotor.setRamp(msUp, msDown: variant): TEv3TachoMotor;
begin
 result:=self;
 setRampUp(msUp);
 setRampDown(msDown);
end;

//--------------------------
procedure Done;
  var ev:TEv3Device;
      i,j:TEv3Port;
 begin
  for i:=outA to outD do
   for j:=outA to outD do begin
    if assigned(ev3Rules[i,j]) then freeAndNil(ev3Rules[i,j]);
    if assigned(ev3MRules[i,j]) then freeAndNil(ev3MRules[i,j]);
   end;

  for ev in ev3DeviceFactory do begin
    case ev.Device of
     ev3tLTachoMotor,ev3tMTachoMotor: begin
//      writeln('+++');
//      writeln(strerror(fpgeterrno));
//      ev.Attrs[ev3acommand].sData:=ev3mCommands[cReset];
//      writeln(ev3mCommands[cReset]);
//      ev.Attrs[ev3acommand].hardPutData(ev3mCommands[cReset]);
//      writeln(strerror(fpgeterrno));
      ev.Attrs[ev3astop_action].sData:=ev3mStopActions[saCoast];
      ev.Attrs[ev3acommand].sData:=ev3mCommands[cStop];
     // WD('Stoped');

//      writeln('----');
    //  readln;
     end;
    end;
  end;
 end;

initialization
 pushDone(@Done);
end.

